h1  {
    margin-top: 40px;
    margin-bottom: 30px;
    padding-bottom: 20px;
    border-bottom: 2px solid black;
    width: 100%;
}

h2  {
    margin-top: 30px;
    margin-bottom: 20px;
    padding-bottom: 15px;
    border-bottom: 1px solid black;
    width: 75%;
}

h3  {
    margin-top: 15px;
    margin-bottom: 10px;
    padding-bottom: 0;
}

.highlight {
    background: #e5f5f5;
    margin-top: 15pt;
}